@@ -4,7 +4,7 @@ module DBusTests.Generation where
 
 import           DBus.Client
 import qualified DBus.Internal.Types as T
-import qualified DBus.Introspection as I
+import qualified DBus.Introspection.Types as I
 import           Data.Int
 import           Data.Map as M
 
@@ -26,6 +26,11 @@ testInterface :: Interface
 testInterface =
   defaultInterface { interfaceMethods =
                        [autoMethod "SampleMethod1" sampleMethod1]
+                   , interfaceProperties =
+                     [autoProperty "SampleWriteProperty"
+                                     (Just $ return (1 :: Int32))
+                                     (Just $ const $ return ())
+                     ]
                    , interfaceName = "org.TestInterface"
                    , interfaceSignals = testSignals
                    }
